import lotus.domino.NotesException;

import javax.management.*;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXServiceURL;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.lang.management.ManagementFactory;
import java.rmi.registry.LocateRegistry;
import java.util.*;

public class JmxServer extends JAddinThread {
    private JMXConnectorServer svr = null;
    private JMXDominoStatistic statObject = null;

    // This is the main entry point. When this method returns, the add-in terminates.
    public void addinStart() {

        logMessage("Started with parameters " + getAddinParameters());

        try {
            logMessage("Running on " + dbGetSession().getNotesVersion());

            int port = 1234;

            Map<String, Object> environment = new HashMap<>();

            LocateRegistry.createRegistry(port);
            MBeanServer mbs = ManagementFactory.getPlatformMBeanServer();
            JMXServiceURL url = new JMXServiceURL("service:jmx:rmi://localhost/jndi/rmi://localhost:" + port + "/jmxrmi");
            svr = JMXConnectorServerFactory.newJMXConnectorServer(url, environment, mbs);

            svr.start();

            registerNewMBean();

        } catch (Exception e) {
            logMessage("Error: " + e.getMessage());
        }

        ArrayList<String> statistics = new ArrayList<>();
        statistics.add("Platform.Memory.*");
        statistics.add("Database.Database.*");

        // Main add-in loop ...
        while (true) {
            //logMessage("User code is executing...");
            waitMilliSeconds(10000L);
            if (statObject != null) {
                statObject.update(getStatistics(statistics));
            }
        }
    }

    // This method is called asynchronously by the JAddin framework when the
    // command 'Quit' or 'Exit' is entered or at Domino server shutdown. Here
    // you may signal the addinStart() method to terminate and to perform any cleanup.
    public void addinStop() {
        logMessage("Termination in progress");

        if (svr != null) {
            try {
                svr.stop();
            } catch (IOException e) {
                logMessage("Error svr stopping: " + e.getMessage());
            }
        }
    }

    // This method is called asynchronously by the JAddin framework for any
    // console command entered. It should be executed as quickly as possible
    // to avoid any main Domino message queue delays.
    public void addinCommand(String command) {
        logMessage("You have entered the command " + command);
    }

    private void registerNewMBean() throws NotCompliantMBeanException, InstanceAlreadyExistsException, MBeanRegistrationException, MalformedObjectNameException {
        statObject = new JMXDominoStatistic();
        MBeanServer mbs = ManagementFactory.getPlatformMBeanServer();

        ObjectName nameStat = new ObjectName("jmx-domino-statistic:type=JmxDominoStatistic");
        mbs.registerMBean(statObject, nameStat);
        System.out.println("MBean registered: " + nameStat.getCanonicalName());
    }

    /**
     * Get all Domino server statistics and create property object
     *
     * @return Property All Domino server statistics
     */
    Map<String, String> getStatistics(List<String> pStatistics) {
        Map<String, String> statValues = new HashMap<>();

        BufferedReader reader;
        String statisticLine;
        String statisticValues;

        String pServerName = "";

        try {
            pServerName = dbGetSession().getServerName();
        } catch (NotesException e) {
            logMessage("Error: dbGetSession().getServerName(): " + e.getMessage());
        }

        logDebug("Method getStatistics() called");

        // Read all defined statistic values
        for (String pStatistic : pStatistics) {

            logDebug("Reading Domino server statistics <" + pStatistic + "> from server " + pServerName);

            // Get statistic values
            try {
                // Undocumented: If the server command starts with "!", the response do not show up in log.nsf
                statisticValues = dbGetSession().sendConsoleCommand(pServerName, "!Show Statistic " + pStatistic);
            } catch (NotesException e) {
                logMessage("Error: Unable to retrieve statistics from Domino server " + pServerName + ": " + e.text);
                return null;
            }

            // Read the string with statistics per line and add it to the property
            try {

                reader = new BufferedReader(new StringReader(statisticValues));

                while ((statisticLine = reader.readLine()) != null) {

                    statisticLine = statisticLine.trim();

                    if (statisticLine.length() == 0)
                        continue;

                    int index = statisticLine.indexOf('=');

                    // Skip invalid entry
                    if (index == -1)
                        continue;

                    // Add the statistic name and value to the property
                    statValues.put(statisticLine.substring(0, index).trim(), statisticLine.substring(index + 1).trim());
                }

                reader.close();

            } catch (IOException e) {
                logMessage("Error: Unable to interpret Domino server statistics <" + pStatistic + ">: " + e.getMessage());
                return null;
            }
        }

        // Add timestamp (number of seconds) to the properties

        logDebug("Number of Domino server statistics read: " + statValues.size());

        return statValues;
    }
}